@@ -803,6 +803,29 @@ public:
         eff_cols_ = find_effective_columns(tmatrix_);        
     }
 
+	/// Transpose array of shorts
+	///
+	void transpose(const GT* BMRESTRICT garr,
+		           unsigned garr_size,
+				   BT* BMRESTRICT tmp_block)
+	{
+		BM_ASSERT(garr_size);
+
+		bit_block_set(tmp_block, 0);
+		::memcpy(tmp_block, garr, sizeof(GT)*garr_size);
+
+        const unsigned arr_size = BLOCK_SIZE * sizeof(unsigned) / sizeof(GT);
+        BM_ASSERT(sizeof(tmatrix_.value) == tmatrix_type::n_columns * 
+                                            tmatrix_type::n_rows * sizeof(GT));
+        // transpose
+        vect_bit_transpose<GT, tmatrix_type::n_rows, tmatrix_type::n_columns>
+                           ((GT*)tmp_block, arr_size, tmatrix_.value);
+
+        // calculate number of non-zero columns
+        eff_cols_ = find_effective_columns(tmatrix_);        
+
+	}
+
     void compute_distance_matrix()
     {
         tmatrix_distance<typename tmatrix_type::value_type, 
